((nil . ((tab-width . 8)
         (fill-column . 70)))
 (change-log-mode . ((add-log-time-zone-rule . t)
		     (fill-column . 74)
		     (bug-reference-url-format . "https://debbugs.gnu.org/%s")
		     (mode . bug-reference)))
 (emacs-lisp-mode . ((bug-reference-url-format . "https://debbugs.gnu.org/%s")
		     (mode . bug-reference-prog)))
 (diff-mode . ((mode . whitespace))))
